@@ -1,9 +1,5 @@
-# -*- coding: utf-8 -*-
-
 """Define the singleton meta class."""
 
-from __future__ import absolute_import
-
 
 class Singleton(type):
     """Implementation of the singleton pattern as a meta class."""
